# -*- coding: utf-8 -*-
from typing import Mapping

from pydantic.json import pydantic_encoder
from ruamel.yaml import YAML
from ruamel.yaml.scalarstring import LiteralScalarString

yaml = YAML()
yaml.default_flow_style = False
yaml.width = 120  # type: ignore
yaml.sequence_dash_offset = 2


def encode_for_yaml(obj):
    if isinstance(obj, Mapping):
        return {k: encode_for_yaml(v) for k, v in obj.items()}
    if isinstance(obj, (list, tuple)):
        return [encode_for_yaml(v) for v in obj]
    try:
        obj = pydantic_encoder(obj)
    except TypeError:
        pass
    if isinstance(obj, str) and "\n" in obj:
        return LiteralScalarString(obj)
    return obj



# -*- coding: utf-8 -*-
from __future__ import annotations

import json
from io import StringIO

import dotenv
from ruamel.yaml import YAMLError

from ._yaml import yaml

SSM_PARAMETER_SUBSTITUTION = (
    ("{{", "ʃ"),  # U+0283	ʃ	ca 83	LATIN SMALL LETTER ESH
    ("}}", "ʅ"),  # U+0285	ʅ	ca 85	LATIN SMALL LETTER SQUAT REVERSED ESH
)


def ssm_curly_to_special(val):
    for sub_set in SSM_PARAMETER_SUBSTITUTION:
        val = val.replace(*sub_set)
    return val


def ssm_special_to_curly(val):
    for sub_set in SSM_PARAMETER_SUBSTITUTION:
        val = val.replace(sub_set[1], sub_set[0])
    return val


def lazy_dict(value):
    """First try as json, then try as yaml, and then try as env file"""
    try:
        return json.loads(value)
    except json.decoder.JSONDecodeError:
        pass
    try:
        return yaml.load(value)
    except YAMLError:
        pass
    return dict(dotenv.dotenv_values(stream=StringIO(value)))



# -*- coding: utf-8 -*-
from __future__ import annotations

from pathlib import PurePath, _PosixFlavour  # type:ignore
from urllib.parse import quote_from_bytes as urlquote_from_bytes

import boto3


class _SSMFlavour(_PosixFlavour):
    is_supported = bool(boto3)

    def parse_parts(self, parts):
        drv, root, parsed = super().parse_parts(parts)
        for part in parsed[1:]:
            if part == "..":
                index = parsed.index(part)
                parsed.pop(index - 1)
                parsed.remove(part)
        return drv, root, parsed

    def make_uri(self, path):
        # We represent the path using the local filesystem encoding,
        # for portability to other applications.
        bpath = bytes(path)
        return "ssm:/" + urlquote_from_bytes(bpath)


_ssm_flavour = _SSMFlavour()


class PureSSMPath(PurePath):
    @classmethod
    def _from_parts(cls, args, *_args, **_kwargs):
        """
        SSM has no concept of a relative path, but root paths don't need the forward slash
        """
        if len(args) > 0:
            if not args[0].startswith("/"):
                args = ["/"] + list(args)
        else:
            args = ["/"]
        return super()._from_parts(args, *_args, **_kwargs)

    _flavour = _ssm_flavour

    @classmethod
    def from_uri(cls, uri):
        if not uri.startswith("ssm://"):
            raise ValueError("Provided uri doesn't seem to be an SSM URI!")
        return cls(uri[5:])



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ssm_parameter_config']

package_data = \
{'': ['*']}

install_requires = \
['boto3',
 'botocore',
 'click>=7.0',
 'pydantic',
 'python-dotenv',
 'ruamel.yaml',
 'signed-pickle>=1']

entry_points = \
{'console_scripts': ['ssm-config-transformer = '
                     'ssm_parameter_config.ssm_writer:cli']}

setup_kwargs = {
    'name': 'ssm-parameter-config',
    'version': '1.1.3',
    'description': 'Description',
    'long_description': '# Ssm Parameter Config\n\n[![Checks][checks-shield]][checks-url]\n[![Codecov][codecov-shield]][codecov-url]\n\n\n\n[codecov-shield]: https://img.shields.io/codecov/c/github/mumblepins/ssm-parameter-config\n[codecov-url]: https://app.codecov.io/gh/mumblepins/ssm-parameter-config\n\n[checks-shield]: https://img.shields.io/github/workflow/status/mumblepins/ssm-parameter-config/Python%20Publish?style=flat-square\n[checks-url]: https://github.com/mumblepins/ssm-parameter-config/actions/workflows/python-publish.yml\n',
    'author': 'Daniel Sullivan',
    'author_email': 'daniel.j.sullivan@state.mn.us',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/mumblepins/ssm-parameter-config/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
from __future__ import annotations

import io
import os
import shlex
from pathlib import Path, PurePath
from typing import Any, Dict, Iterable, Mapping, Optional, Type, Union

from pydantic import BaseSettings, ValidationError, parse_obj_as
from pydantic.env_settings import DotenvType, SettingsSourceCallable, env_file_sentinel

from ._yaml import encode_for_yaml, yaml
from .ssm_parameter import SSMParameter
from .utils import lazy_dict

StrPath = Union[str, os.PathLike, PurePath]


class LocalSSMSettings:
    __slots__ = ("local_ssm_path",)

    def __init__(self, local_ssm_path: Optional[StrPath]):
        self.local_ssm_path: Optional[StrPath] = local_ssm_path

    def __call__(self, settings: BaseSettings) -> Dict[str, Any]:
        settings_path: StrPath
        if self.local_ssm_path is not None:
            settings_path = self.local_ssm_path
        elif "LOCAL_SSM_SETTINGS_PATH" in os.environ:
            settings_path = os.environ["LOCAL_SSM_SETTINGS_PATH"]
        elif hasattr(settings.__config__, "local_ssm_settings_path"):
            settings_path = settings.__config__.local_ssm_settings_path
        else:
            return {}

        spath = Path(settings_path).expanduser()
        if not spath.exists():
            return {}

        pdict = lazy_dict(spath.read_text(encoding="utf8"))

        if settings.__config__.case_sensitive:
            return pdict

        return {k.lower(): v for k, v in pdict.items()}

    def __repr__(self) -> str:
        return f"LocalSSMSettings(yaml_file={self.local_ssm_path!r})"


class AwsSSMSettings:
    __slots__ = ("ssm_path",)

    def __init__(self, ssm_path: Optional[str]):
        self.ssm_path: Optional[str] = ssm_path

    def __call__(self, settings: BaseSettings) -> Dict[str, Any]:
        settings_path: str
        if self.ssm_path is not None:
            settings_path = self.ssm_path
        elif "AWS_SSM_SETTINGS_PATH" in os.environ:
            settings_path = os.environ["AWS_SSM_SETTINGS_PATH"]
        elif hasattr(settings.__config__, "aws_ssm_settings_path"):
            settings_path = settings.__config__.aws_ssm_settings_path
        else:
            return {}

        param = SSMParameter.get_parameter(settings_path)
        if param.value == "":
            return {}
        pdict = param.lazy_dict()
        pdict["ssm_parameter"] = param
        if settings.__config__.case_sensitive:
            return pdict

        return {k.lower(): v for k, v in pdict.items()}

    def __repr__(self) -> str:
        return f"LocalSSMSettings(yaml_file={self.ssm_path!r})"


_ssm_config_classes: list[Type[SSMConfig]] = []


class SSMConfig(BaseSettings):
    def __init__(  # pylint: disable=no-self-argument
        __pydantic_self__,
        *,
        _env_file: Optional[DotenvType] = env_file_sentinel,
        _env_file_encoding: Optional[str] = None,
        _env_nested_delimiter: Optional[str] = None,
        _secrets_dir: Optional[StrPath] = None,
        _local_ssm_path: Optional[StrPath] = None,
        _aws_ssm_path: Optional[str] = None,
        **values: Any,
    ) -> None:
        if _local_ssm_path is not None:
            values["_local_ssm_path"] = _local_ssm_path

        if _aws_ssm_path is not None:
            values["_aws_ssm_path"] = _aws_ssm_path
        super().__init__(_env_file, _env_file_encoding, _env_nested_delimiter, _secrets_dir, **values)

    ssm_parameter: Optional[SSMParameter] = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        _ssm_config_classes.append(cls)

    class Config:
        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ):
            # this is an ugly way to store these special variables without overriding
            # the whole BaseSettings._build_values function
            local_source = LocalSSMSettings(
                local_ssm_path=init_settings.init_kwargs.pop("_local_ssm_path", None)  # type: ignore[attr-defined]
            )
            aws_source = AwsSSMSettings(
                ssm_path=init_settings.init_kwargs.pop("_aws_ssm_path", None)  # type: ignore[attr-defined]
            )
            return (
                init_settings,
                local_source,
                aws_source,
                env_settings,
                file_secret_settings,
            )

    def export(self, exp_format="yaml", exclude_ssm=True, ssm_format=False, **dump_kwargs):
        dict_args = {"exclude_none": True, "by_alias": True, "exclude_defaults": True}
        if exclude_ssm:
            dict_args["exclude"] = {"ssm_parameter"}
        if exp_format == "json":
            kwargs = {"separators": (",", ":")}
            kwargs.update(dump_kwargs)
            output = self.json(**kwargs, **dict_args)
        elif exp_format == "yaml":
            kwargs = {}
            kwargs.update(dump_kwargs)
            out = io.StringIO()
            yaml.dump(encode_for_yaml(self.dict(**dict_args)), out, **kwargs)
            output = out.getvalue()
        elif exp_format == "env":
            exp = []
            for k, v in self.dict(**dict_args).items():
                if isinstance(v, (Mapping, Iterable)) and not isinstance(v, (str, bytes, bytearray)):
                    v = self.__config__.json_dumps(v, default=self.__json_encoder__)
                v = shlex.quote(v)
                exp.append(f"{k.upper()}={v}")
            exp.append("")
            output = "\n".join(exp)
        else:
            raise ValueError(f"Format {exp_format} not supported.")
        if ssm_format:
            return SSMParameter.get_parameter_value(output)
        return output

    @classmethod
    def from_object(cls, obj: dict[str, Any]):
        for c in reversed(_ssm_config_classes):
            try:
                new_cls = parse_obj_as(c, obj)
                break
            except ValidationError:
                pass
        else:
            raise ValueError(f"Could not parse {obj} as any of {_ssm_config_classes!r}")
        return new_cls

    @classmethod
    def from_file(cls, file: StrPath) -> SSMConfig:
        for c in reversed(_ssm_config_classes):
            try:
                new_cls = c(_local_ssm_path=file)
                break
            except ValidationError:
                pass
        else:
            raise ValueError(f"Could not parse {file!s} as any of {_ssm_config_classes!r}")
        return new_cls

    @classmethod
    def from_parameter(cls, parameter: SSMParameter) -> SSMConfig:
        ld = parameter.lazy_dict()
        new_cls = cls.from_object(ld)
        new_cls.ssm_parameter = parameter
        return new_cls

    def _write_config_env(self):
        if isinstance(self.__config__.env_file, (list, tuple)):
            efile = self.__config__.env_file[0]
        else:
            efile = self.__config__.env_file
        with open(efile, "wt", encoding="utf8") as fh:
            fh.write(self.export("env"))

    def to_parameter(self, exp_format="yaml", ssm_parameter_path: Optional[str] = None, ignore_current: bool = False):
        if ssm_parameter_path is not None:
            if not ignore_current:
                ssm_param = SSMParameter.get_parameter(ssm_parameter_path)
            else:
                ssm_param = SSMParameter(Name=ssm_parameter_path, Value="")
        else:
            ssm_param = self.ssm_parameter
        # if ssm_param.value is None or ssm_param.value == "":
        ssm_param.value = self.export(exp_format)
        if ssm_param != self.ssm_parameter:
            self.ssm_parameter = ssm_param
        return ssm_param

    def _write_config_ssm(self, exp_format, ssm_parameter_path=None, as_cli_input: bool = False):
        kwargs = {}
        if as_cli_input:
            kwargs = {"ignore_current": True}
        ssm_param = self.to_parameter(exp_format=exp_format, ssm_parameter_path=ssm_parameter_path, **kwargs)
        return ssm_param.put_parameter(as_cli_input=as_cli_input)

    def _write_config_local(self, exp_format, path):
        val = self.export(exp_format)
        with open(path, "wt", encoding="utf8") as fh:
            fh.write(val)

    def write_config(
        self,
        exp_format="yaml",
        ssm_parameter_path=None,
        local_path=None,
        as_cli_input: bool = False,
    ):
        if ssm_parameter_path is not None or (self.ssm_parameter is not None and exp_format != "env"):
            return self._write_config_ssm(exp_format, ssm_parameter_path=ssm_parameter_path, as_cli_input=as_cli_input)
        if hasattr(self.__config__, "local_settings_path") or local_path is not None:
            if local_path is not None:
                output = Path(local_path)
            else:
                output = Path(self.__config__.local_settings_path)  # type:ignore
            output = output.expanduser()
            self._write_config_local(exp_format, path=output)
        elif hasattr(self.__config__, "env_file"):
            self._write_config_env()
        else:
            raise ValueError("No SSM parameter (path) or env file defined.")
        return None



# -*- coding: utf-8 -*-
# pylint: disable=too-many-arguments,too-many-locals,no-value-for-parameter
import importlib
import json
import random
import string
import sys

import click

from ssm_parameter_config import SSMConfig, SSMParameter  # nodep


@click.command()
@click.argument("input_file", type=click.Path(file_okay=True, dir_okay=False, path_type=str, allow_dash=True))
@click.argument("output", type=click.File("w"), default="-")
@click.option("--ssm-config-format", type=click.Choice(["json", "yaml"]), default="yaml")
@click.option("--output-format", type=click.Choice(["json", "shell"]), default="json")
@click.option("--as-config", is_flag=True)
@click.option("--extra-import", type=click.STRING, multiple=True)
@click.option("--extra-import-path", type=click.Path(exists=True, file_okay=False, dir_okay=True), multiple=True)
@click.option("--push-to-aws", is_flag=True)
@click.option("--parameter-name", type=click.STRING, default=None)
@click.option("--parameter-base-name", type=click.STRING, default="")
@click.option("--pretty-print-json", is_flag=True, default=True)
def cli(
    input_file: str,
    output,
    ssm_config_format,
    push_to_aws,
    output_format,
    as_config,
    extra_import,
    extra_import_path,
    parameter_name,
    parameter_base_name,
    pretty_print_json,
):
    if parameter_name is None:
        if input_file == "-":
            raise click.UsageError("Need to specify --parameter_name if reading from STDIN")
        parameter_name = input_file.split("/")[-1]
    parameter_full_name = parameter_base_name + parameter_name
    output.write(
        main(
            input_file,
            ssm_config_format,
            push_to_aws,
            output_format,
            as_config,
            extra_import,
            extra_import_path,
            parameter_full_name,
            pretty_print_json,
        )
    )


def main(
    input_file,
    ssm_config_format,
    push_to_aws,
    output_format,
    as_config,
    extra_import,
    extra_import_path,
    parameter_full_name,
    pretty_print_json,
):
    if as_config:
        for eip in extra_import_path:
            sys.path.append(str(eip))
        for ei in extra_import:
            importlib.import_module(ei)
        input_ssm_cfg = SSMConfig.from_file(input_file)
        input_ssm = input_ssm_cfg.to_parameter(
            exp_format=ssm_config_format, ssm_parameter_path=parameter_full_name, ignore_current=True
        )
    else:
        with click.open_file(input_file, "r", encoding="utf8") as fh:
            input_ssm = SSMParameter(Name=parameter_full_name, Value=fh.read())
    out = input_ssm.put_parameter(as_cli_input=not push_to_aws)
    if push_to_aws:
        return ""
    output = ""
    if output_format == "shell":
        var_name = "".join(random.choices(string.ascii_letters, k=10))  # nosec B311
        output += f"read -r -d '' {var_name} <<'EOF_{var_name}'\n"
    if pretty_print_json:
        output += json.dumps(out, indent=2)
    else:
        output += json.dumps(out)
    if output_format == "shell":
        output += f"\nEOF_{var_name}\n\n"
        output += f'aws ssm put-parameter --cli-input-json "${var_name}"\n'
    return output


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
from __future__ import annotations

import base64
import logging
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Any, Iterator, Optional

import boto3
from botocore.client import BaseClient
from pydantic import BaseModel, PrivateAttr, parse_obj_as
from pydantic.utils import to_camel
from signed_pickle import DumperSigner

from .ssm_path import PureSSMPath
from .utils import lazy_dict, ssm_curly_to_special, ssm_special_to_curly

if TYPE_CHECKING:
    from .ssm_config import SSMConfig

# since SSM Parameters can't have {{ }} in them, we substitute these values

logger = logging.getLogger()


class Tag(BaseModel):
    key: str
    value: str

    class Config:
        alias_generator = to_camel


class SSMType(str, Enum):
    string = "String"
    string_list = "StringList"
    secure_string = "SecureString"


class SSMTier(str, Enum):
    standard = "Standard"
    advanced = "Advanced"
    intelligent_tiering = "Intelligent-Tiering"


class SSMDataType(str, Enum):
    text = "text"
    ec2_image = "aws:ec2:image"
    ssm_integration = "aws:ssm:integration"


class SSMPath(BaseModel):
    """kind of like a pathlib path, but not quite"""

    name: str
    _listed: bool = PrivateAttr(False)
    _children: dict[str, SSMPath] = PrivateAttr({})
    _aws_client_kwargs: dict[str, Any] = PrivateAttr(default={})

    def __getitem__(self, item):
        if isinstance(item, tuple):
            this = self
            for i in item:
                this = this[i]
            return this
        try:
            return getattr(self, item)
        except AttributeError:
            pass
        self._fetch_children()
        if item in self._children:
            return self._children[item]
        nc = SSMPath(name=str(self.path / item))
        self._children[item] = nc
        return nc

    def __setitem__(self, item, value):
        if isinstance(item, tuple):
            this = self
            for i in item[:-1]:
                this = this[i]
            this[item[-1]] = value
            return
        try:
            setattr(self, item, value)
            return
        except ValueError:
            pass
        self._children[item] = value

    @property
    def path(self):
        return PureSSMPath(self.name)

    def fetch_parameters(self, path):
        ssm = self.ssm_client
        get_params_pager = ssm.get_paginator("get_parameters_by_path")
        desc_params_pager = ssm.get_paginator("describe_parameters")
        params = {}

        # fetch descriptions
        for page in desc_params_pager.paginate(
            ParameterFilters=[{"Key": "Name", "Option": "BeginsWith", "Values": [path]}]
        ):
            for p in page["Parameters"]:
                params[p["Name"]] = p
        # fetch values
        for page in get_params_pager.paginate(Path=path, Recursive=True):
            for p in page["Parameters"]:
                params[p["Name"]].update(p)

        for p in params.values():
            p["Value"] = ssm_special_to_curly(p["Value"])
            param = parse_obj_as(SSMParameter, p)

            rel_path = param.path.relative_to(self.path).parts
            self[rel_path] = param
            parts = []
            for part in rel_path[:-1]:
                parts.append(part)
                self[tuple(parts)]._listed = True  # pylint:disable=protected-access
        self._listed = True

    def _fetch_children(self):
        if self._listed:
            return

        self._listed = True
        ssm = self.ssm_client
        logger.info("Getting children for %s", self.name)
        get_params_pager = ssm.get_paginator("get_parameters_by_path")
        desc_params_pager = ssm.get_paginator("describe_parameters")
        params = {}

        # fetch descriptions
        for page in desc_params_pager.paginate(
            ParameterFilters=[{"Key": "Name", "Option": "BeginsWith", "Values": [self.name]}]
        ):
            for p in page["Parameters"]:
                params[p["Name"]] = p
        # fetch values
        for page in get_params_pager.paginate(Path=self.name, Recursive=True):
            for p in page["Parameters"]:
                params[p["Name"]].update(p)

        for p in params.values():
            p["Value"] = ssm_special_to_curly(p["Value"])
            param = parse_obj_as(SSMParameter, p)
            rel_path = param.path.relative_to(self.path).parts
            self[rel_path] = param
            parts = []
            for part in rel_path[:-1]:
                parts.append(part)
                self[tuple(parts)]._listed = True  # pylint:disable=protected-access

    def iterdir(self) -> Iterator[SSMPath]:
        self._fetch_children()

        yield from self._children.values()

    def is_dir(self):
        return True

    def is_file(self):
        return not self.is_dir()

    def set_aws_client_kwargs(self, **kwargs):
        self._aws_client_kwargs = kwargs

    @property
    def ssm_client(self):
        return boto3.client("ssm", **self._aws_client_kwargs)


class SSMParameter(SSMPath):
    # name (from SSMPATH) # from get_param..
    description: Optional[str] = None  # from describe_param
    value: Optional[str] = None  # from get_param..
    type: SSMType = SSMType.string  # from get_param..
    key_id: Optional[str] = None  # from describe_param
    allowed_pattern: Optional[str] = None  # from describe_param
    version: Optional[int] = None  # from get_param..
    last_modified_date: Optional[datetime] = None  # from get_param..
    tier: SSMTier = SSMTier.standard  # from describe_param
    data_type: SSMDataType = SSMDataType.text  # from get_param..
    tags: list[Tag] = []  # from list_tags
    _decoded_value: Any = PrivateAttr(default=None)
    _got_tags: bool = PrivateAttr(default=False)

    class Config:
        alias_generator = to_camel
        exclude_none = True

    def __getattribute__(self, item):
        # if item == "value" and super().__getattribute__("value") is None:
        #     # lazily fetch value upon first read
        #     self._fetch_value()
        if item == "tags" and not super().__getattribute__("_got_tags"):
            self._fetch_tags()
        return super().__getattribute__(item)

    def is_dir(self):
        return False

    def as_ssm_config(self) -> SSMConfig:
        from .ssm_config import SSMConfig  # pylint:disable=import-outside-toplevel

        return SSMConfig.from_parameter(self)

    # def _fetch_value(self):
    #     ssm: BaseClient = self.ssm_client
    #     try:
    #         param = ssm.get_parameter(Name=self.name)["Parameter"]
    #     except (IndexError, ssm.exceptions.ParameterNotFound):
    #         self.__setattr__("value", "")
    #         return
    #     new_param = parse_obj_as(SSMParameter, param)
    #     new_vals = new_param.dict(exclude_unset=True, exclude_none=True, exclude_defaults=True)
    #     # new_vals['value'] = new_vals['value'].
    #     for k, v in new_vals.items():
    #         self.__setattr__(k, v)

    def _fetch_tags(self):
        ssm: BaseClient = self.ssm_client
        try:
            tags = ssm.list_tags_for_resource(ResourceType="Parameter", ResourceId=self.name)
            self.tags = parse_obj_as(list[Tag], tags["TagList"])
            self._got_tags = True
        except ssm.exceptions.ParameterNotFound:
            pass

    @property
    def decoded_value(self):
        if self._decoded_value is None:
            try:
                decoded_val = base64.b64decode(self.value.strip(), validate=True)
                self._decoded_value = DumperSigner.load(decoded_val)[0].decode("utf8")
            except ValueError:
                self._decoded_value = self.value
        return self._decoded_value

    def lazy_dict(self):
        return lazy_dict(self.decoded_value)

    @classmethod
    def get_parameter(cls, name: str, default_value: str = ""):
        ssm = boto3.client("ssm")
        try:
            param = ssm.get_parameter(Name=name)["Parameter"]
            param.update(ssm.describe_parameters(ParameterFilters=[{"Key": "Name", "Values": [name]}])["Parameters"][0])
            tags = ssm.list_tags_for_resource(ResourceType="Parameter", ResourceId=name)
            param["Tags"] = tags["TagList"]
            param["Value"] = ssm_special_to_curly(param["Value"])
            return parse_obj_as(cls, param)
        except (IndexError, ssm.exceptions.ParameterNotFound):
            pass
        return cls(Name=name, Value=default_value)

    @staticmethod
    def get_parameter_value(val):
        if len(val) > 4096:
            val = base64.b64encode(DumperSigner().dump(val.encode("utf8"))).decode("utf8")

        return ssm_curly_to_special(val)

    def put_parameter(self, new_value=None, as_cli_input: bool = False):
        val = self.get_parameter_value(new_value or self.value)
        kwargs = self.dict(
            exclude_none=True,
            # exclude_defaults=True,
            by_alias=True,
            exclude={"version", "last_modified_date", "value", "tags"},
            # include={"type"},
        )

        kwargs["Value"] = val
        kwargs["Overwrite"] = True
        if as_cli_input:
            return kwargs
        ssm = boto3.client("ssm")
        ssm.put_parameter(**kwargs)
        return None
